"""Common code shared between modules.

This module is intended to be imported with 'from ... import *' semantics and
provides an __all__ specification for this purpose.

"""

import ctypes
import math

__all__ = ['floatround', 'raw_values', '_ease', '_wibble', '_pop', 'bytwos', 'pairs', 'cyclicpairs', 'disjoint_pairs', 'formatscore', 'formatdate', 'cached', 'ghostify_svg_name']

def floatround(x):
    return int(x * 65536) / 65536.0
    
def raw_values(v):
    cv = (ctypes.c_double * 2)(*v)
    return ctypes.cast(cv, ctypes.POINTER(ctypes.c_ubyte))[:16]
    
def _ease(t):
    sgn = [-1, 1][t > 0]
    return sgn * (3 * abs(t) ** 2 - 2 * abs(t) ** 3)

def _wibble(t, k=5):
    return _ease(1 - abs(t)) * math.sin(2 * math.pi * k * t)

def _pop(t):
    return t ** 2

def bytwos(seq):
    for ii in xrange(len(seq) // 2):
        yield (seq[2*ii], seq[2*ii + 1])
def pairs(seq):
    for ii in xrange(len(seq) - 1):
        p1 = seq[ii]
        p2 = seq[ii+1]
        yield (p1, p2)

def cyclicpairs(seq):
    for ii in xrange(len(seq)):
        p1 = seq[ii]
        p2 = seq[ii+1] if ii < len(seq) - 1 else seq[0]
        yield (p1, p2)
        
def disjoint_pairs(iterable):
    seq = list(iterable)
    for ii in xrange(len(seq) - 1):
        for jj in xrange(len(seq) - ii - 1):
            p1 = seq[ii]
            p2 = seq[ii + jj + 1]
            yield (p1, p2)

def parsescore(ticks):
    from constants import TICK_RATE
    seconds, fraction = divmod(ticks / (100 * TICK_RATE), 1.0)
    minutes = int(seconds / 60.0)
    seconds = int(seconds - minutes * 60)
    hundredths = int(fraction * 100.0)
    return minutes, seconds, hundredths

def formatscore(ticks):
    return u"%d:%02d:%02d" % parsescore(ticks)

def formatdate(date):
    from datetime import datetime
    import time
    dt = time.time() - date
    if dt < 1: return u"in the future"
    days = dt // 60 // 60 // 24
    hours = dt // 60 // 60 - days * 24
    minutes = dt // 60 - hours * 60 - days * 24 * 60
    seconds = dt // 1 - minutes * 60 - hours * 60 * 60 - days * 24 * 60 * 60
    result = [None, "ago"]
    for name in ["day", "hour", "minute", "second"]:
        v = locals()[name + "s"]
        if v > 0:
            result[0] = "%d %s" % (v, name)
            if v > 1:
                result[0] += "s"
            return unicode(" ".join(result))

def cached(func):
    """Decorate a function as a caching property.

    :Parameters:
        `func` : function
            The getter function to decorate.

    """
    cached_name = "_cached_%s" % func.func_name

    # The keywords 'getattr' and 'cached_name' are used to optimise the common
    # case (return cached value) by bringing the used names to local scope.
    def fget(self, getattr=getattr, cached_name=cached_name):
        try:
            return getattr(self, cached_name)
        except AttributeError:
            value = func(self)
            setattr(self, cached_name, value)
            return value

    def fset(self, value):
        setattr(self, cached_name, value)

    def fdel(self):
        try:
            delattr(self, cached_name)
        except AttributeError:
            pass

    fget.func_name = "get_" + func.func_name
    fset.func_name = "set_" + func.func_name
    fdel.func_name = "del_" + func.func_name

    return property(fget, fset, fdel, doc=func.func_doc)
    
def ghostify_svg_name(name):
    if name in ['frenchship.svg', 'italianship.svg', 'irishship.svg']:
        return 'tricolourghost.svg'
    if name == 'usaship.svg':
        return 'bwship.svg'
    return 'ghost' + name
