"""Interface components.

"""

from __future__ import division

from pyglet import graphics
from pyglet.event import EVENT_HANDLED
from pyglet.event import EVENT_UNHANDLED
from pyglet.text import caret
from pyglet.text import document
from pyglet.text import layout
from pyglet.window import key
from pyglet.window import mouse
from pyglet.gl import *

import config
from common import *
from constants import *


def make_callback(callback):
    if isinstance(callback, tuple):
        func = callback[0]
        args = callback[1:]
        return func(*args)
    return callback()


class _Layout(layout.IncrementalTextLayout):
    """Custom TextLayout subclass.

    """

    def __init__(self, document, interface):
        self._interface = interface
        super(_Layout, self).__init__(document, 0, 0)

    def _update(self):
        self.content_width = self.content_height = 0
        self._interface._clean = False
        self._interface._layout._clean = False
        super(_Layout, self)._update()


class _Document(document.UnformattedDocument):
    """Custom AbstractDocument subclass.

    """

    pass


class _Caret(caret.Caret):
    """Custom Caret subclass.

    """

    def __init__(self, entry):
        super(_Caret, self).__init__(entry._text_layout, color=(255,255,255))
        self._entry = entry
        self._reset()

    def _reset(self):
        self._layout.document.text = self._entry._default
        self.mark = 0
        self.position = len(self._layout.document.text)
        self._update()
        self._next_attributes.clear()

    def on_text(self, text):
        if text == u"\r":
            self._entry._callback(self._layout.document.text)
            self._reset()
            return EVENT_HANDLED
        else:
            if len(self._layout.document.text) < 30:
                return super(_Caret, self).on_text(text)


    def on_mouse_press(self, x, y, btn, mods):
        pass

    def on_mouse_release(self, x, y, btn, mods):
        pass

    def on_mouse_drag(self, x, y, dx, dy, btns, mods):
        pass


class Meta(type):

    def __init__(self, name, bases, dict):

        _kwds = dict.get("_kwds", {})
        _scaled = dict.get("_scaled", [(), ()])

        # Create keyword properties.
        for key, value in _kwds.items():

            # Names.
            getter_name = "get_" + key
            setter_name = "set_" + key
            shadow_key = "_" + key

            # Create property.
            def getter(self, name=getter_name):
                return getattr(self, name)()
            def setter(self, value, name=setter_name):
                return getattr(self, name)(value)
            prop = property(getter, setter)
            setattr(self, key, prop)
            dict[key] = prop

            # Create getter.
            if not hasattr(self, getter_name):
                def getter(self, key=shadow_key):
                    return getattr(self, key)
                getter.func_name = getter_name
                setattr(self, getter_name, getter)
                dict[getter_name] = getter

            # Create setter.
            if not hasattr(self, setter_name):
                def setter(self, value, key=shadow_key):
                    return setattr(self, key, value)
                setter.func_name = setter_name
                setattr(self, setter_name, setter)
                dict[setter_name] = setter

            # Create default.
            setattr(self, shadow_key, value)
            dict[shadow_key] = value

            # Create horizontal scaler.
            if key in self._scaled[0]:

                # Create property.
                def getter(self, key=shadow_key):
                    value = getattr(self, key)
                    return value * self.parent_size[0]
                getter.func_name = "real_" + key
                setattr(self, "real_" + key, cached(getter))

            # Create vertical scaler.
            if key in _scaled[1]:

                # Create property.
                def getter(self, key=shadow_key):
                    value = getattr(self, key)
                    return value * self.parent_size[1]
                getter.func_name = "real_" + key
                setattr(self, "real_" + key, cached(getter))

        self._cached = getattr(bases[0], "_cached", []) + \
                ["real_" + key for list in _scaled for key in list]



class Interface(object):

    # Metaclass.

    __metaclass__ = Meta

    # Keywords.

    _kwds = {
        "xpos" : 0.0,
        "ypos" : 0.0,
        "width" : 1.0,
        "height" : 1.0,
        "halign" : "center",
        "valign" : "center",
        "window" : None,
        "layout" : None,
        "padding" : 0.0,
        "margin" : 0.0,
        "background" : None,
        "outline" : None,
        "expand" : False,
        "show" : True,
        "xoff" : 0.0,
        "yoff" : 0.0,
        "aoff" : 0.0,
        "popup" : 1.0,
        "always_popup" : False,
    }

    _scaled = [
        ("xpos", "width"),
        ("ypos", "height", "padding", "margin"),
    ]

    # Symbols.

    _syms = {
        "bottom" : 0.0,
        "middle" : 0.5,
        "top" : 1.0,
        "left" : 0.0,
        "center" : 0.5,
        "right" : 1.0,
    }

    _no_event = False
    _ign_event = False

    # Constructor.

    def __init__(self, **kwds):
        self._clean = False
        self._components = []
        for key, value in kwds.items():
            setattr(self, key, value)

    def add(self, interface_cls, **kwds):
        kwds["layout"] = self
        return interface_cls(**kwds)

    def clear(self):
        for component in self._components:
            component.window = None
        self._components = []

    # Properties.

    def set_window(self, window):
        if self._window is not None and not self._no_event:
            self._window.remove_handlers(self)
        self._window = window
        for component in self._components:
            component.window = window
        if self._window is not None and not self._no_event:
            self._window.push_handlers(self)

    def set_layout(self, layout):
        if self._layout is not None:
            self._layout._components.remove(self)
        self._layout = layout
        if self._layout is not None:
            self._layout._components.append(self)
            self.window = self._layout.window

    # Dimensions.

    def _parent_offset(self):
        """Offset of the parent interface's layout's content corner.

        """
        return self._layout.layout_offset
    parent_offset = property(lambda self: self._parent_offset())

    def _parent_size(self):
        """Dimensions of the parent interface's layout.

        """
        return self._layout.layout_size
    parent_size = property(lambda self: self._parent_size())

    def _layout_offset(self):
        """Offset of the layout's content corner.

        """
        cxoff, cyoff = self.content_offset
        lxoff, lyoff = self._layout.layout_offset
        return lxoff + cxoff, lyoff + cyoff
    layout_offset = property(lambda self: self._layout_offset())

    def _layout_size(self):
        """Dimensions of the layout.

        """
        return self.real_width - 2 * (self.real_padding + self.real_margin), \
               self.real_height - 2 * (self.real_padding + self.real_margin)
    layout_size = property(lambda self: self._layout_size())

    def _content_offset(self):
        """Offset of the interface's content corner.

        """
        return self.real_xpos + self.real_padding + self.real_margin + self.xoff, \
               self.real_ypos + self.real_padding + self.real_margin + self.yoff
    content_offset = property(lambda self: self._content_offset())

    def _content_size(self):
        """Dimensions of the interface's content.

        """
        return self.real_width - 2 * (self.real_padding + self.real_margin), \
               self.real_height - 2 * (self.real_padding + self.real_margin)
        return layout.real_width - 2 * (layout.real_padding + layout.real_margin), \
               layout.real_height - 2 * (layout.real_padding + layout.real_margin)
    content_size = property(lambda self: self._content_size())

    def _box_offset(self):
        """Offset of the interface's box corner.

        """
        xpos, ypos = self.content_offset
        return xpos - self.real_padding, \
               ypos - self.real_padding
    box_offset = property(lambda self: self._box_offset())

    def _box_size(self):
        """Dimensions of the interface's box.

        """
        width, height = self.content_size
        return width + 2 * self.real_padding, \
               height + 2 * self.real_padding
    box_size = property(lambda self: self._box_size())

    def reposition(self, xpos, ypos):
        """Dynamically reposition at layout coordinates.

        """
        self.real_xpos = xpos
        self.real_ypos = ypos

    def resize(self, width, height):
        """Dynamically resize to absolute dimensions.

        """
        self.real_width = width
        self.real_height = height

    def realign(self):
        """Perform dynamic realignment.

        """
        width, height = self.content_size
        lwidth, lheight = self.layout_size
        del self.real_xpos, self.real_ypos
        self.real_xpos += (lwidth - width) * self._syms[self._halign]
        self.real_ypos += (lheight - height) * self._syms[self._valign]

    def organise(self):
        """Instruct components to reposition.

        """
        corners = [(0, 0)]
        for component in self._components:
            x, y = component.content_offset
            w, h = component.content_size
            corners.append((x + w, y))
            corners.append((x, y + h))
        corners = zip(*corners)
        x, y = map(min, corners)
        w, h = map(max, corners)

    # Rendering.

    def clean_content(self):
        """Ensure the content is in a renderable state.

        """

    def make_clean(self):
        """Ensure the layout tree is in a renderable state.

        """
        if not self._clean:
            self.clean_content()
            for component in self._components:
                component.make_clean()
            self.organise()
            self.realign()
            self._clean = True

    def draw(self):
        """Render everything to screen.

        """
        if self.show:
            self.make_clean()
            self.draw_box()
            glPushMatrix()
            xoff, yoff = self.content_offset
            w, h = self.content_size
            glTranslatef(xoff + w / 2, yoff + h / 2, 0.0)
            glRotatef(self.aoff, 0.0, 0.0, 1.0)
            if (self._mouse_in_interface and self._popup != 1.0) or self._always_popup:
                glScalef(self._popup, self._popup, 1.0)
            glTranslatef(-w / 2, -h / 2, 0.0)
            for component in self._components:
                component.draw()
            self.draw_content()
            glPopMatrix()

    def draw_box(self):
        """Render the background box to screen.

        """
        (x, y), (w, h) = self.box_offset, self.box_size
        vdata = ('v2f', [x, y, x + w, y, x + w, y + h, x, y + h])
        if self._background is not None:
            graphics.draw(4, GL_QUADS, vdata, ('c4B', self._background * 4))
        if self._outline is not None:
            graphics.draw(4, GL_LINE_LOOP, vdata, ('c4B', self._outline * 4))

    def draw_content(self):
        """Render the content to screen.

        """
        pass

    # Events.

    def _in_interface(self, x, y):
        pxoff, pyoff = self.parent_offset
        xoff, yoff = self.content_offset
        xoff, yoff = xoff + pxoff, yoff + pyoff
        width, height = self.content_size
        return xoff <= x < xoff + width and \
               yoff <= y < yoff + height

    _mouse_in_interface = False
    def set_mouse(self, x, y):
        old = self._mouse_in_interface
        if self._in_interface(x, y):
            self._mouse_in_interface = new = True
        else:
            self._mouse_in_interface = new = False
        if old and not new:
            self.leave()
        elif new and not old:
            self.enter()

    def on_mouse_release(self, x, y, button, modifiers):
        if not self._ign_event and self.show:
            self.set_mouse(x, y)
            if self._mouse_in_interface:
                if self.click():
                    return EVENT_HANDLED
            return EVENT_UNHANDLED

    def on_mouse_motion(self, x, y, dx, dy):
        if not self._ign_event and self.show:
            self.set_mouse(x, y)

    def enter(self):
        pass

    def leave(self):
        pass

    def click(self):
        pass


class Root(Interface):

    _kwds = {
        "offset" : (0, 0),
        "size" : None,
    }

    def _parent_offset(self):
        return self.offset

    def _parent_size(self):
        if self.size is None:
            return self._window.get_size()
        return self.size

    def _layout_offset(self):
        return self.offset


class Text(Interface):
    """Text display.

    """

    # Keywords.

    _kwds = {
        "font_name" : FONT_NAME,
        "font_size" : 0.08,
        "color" : (255, 255, 255, 255),
        "text" : u""
    }

    _scaled = [
        (),
        ("font_size",),
    ]

    # Constructor.

    def __init__(self, **kwds):
        super(Text, self).__init__(**kwds)
        self._document = _Document(self._text)
        self._text_layout = _Layout(self._document, self)
        self._text_layout.selection_background_color = (0, 0, 0, 80)

    # Properties.

    def set_text(self, text):
        self._text = text
        if hasattr(self, "_document"):
            self._document.text = self._text

    # Dimensions.

    def _content_size(self):
        return self._text_layout.content_width, \
               self._text_layout.content_height

    def reposition(self, xpos, ypos):
        super(Text, self).reposition(xpos, ypos)

    def realign(self):
        super(Text, self).realign()

    def set_color(self, color):
        self._color = color
        self._clean = False

    # Rendering.

    def clean_content(self):
        super(Text, self).clean_content()
        self._text_layout.begin_update()
        # Configure document.
        self._document.set_style(0, 0, {
            "font_name" : self._font_name,
            "font_size" : int(self.real_font_size),
            "color" : self._color,
            "align" : "left",
        })
        # Configure layout.
        w, h = self.layout_size
        self._text_layout.width = w
        self._text_layout.height = h
        self._text_layout.content_valign = "bottom"
        self._text_layout.multiline = True
        self._text_layout.end_update()

    def draw_content(self):
        self._text_layout.draw()



class Link(Text):

    _kwds = {
        "normal" : LINK_NORMAL,
        "hilight" : LINK_HILIGHT,
        "callback" : staticmethod(lambda: None)
    }

    # Events.

    def enter(self):
        self.color = self.hilight
        self._clean = False
        self._layout._clean = False

    def leave(self):
        self.color = self.normal
        self._clean = False
        self._layout._clean = False

    def click(self):
        make_callback(self.callback)
        return True


class CLink(Link):
    def realign(self):
        super(CLink, self).realign()
    def _content_offset(self):
        x, y = super(CLink, self)._content_offset()
        return x - (self.layout_size[0] - self.content_size[0]) / 2, y
    def _in_interface(self, x, y):
        pxoff, pyoff = self.parent_offset
        xoff, yoff = self.content_offset
        xoff += (self.layout_size[0] - self.content_size[0]) / 2
        xoff, yoff = xoff + pxoff, yoff + pyoff
        width, height = self.content_size
        return xoff <= x < xoff + width and \
               yoff <= y < yoff + height
    def clean_content(self):
        super(Link, self).clean_content()
        self._text_layout.begin_update()
        # Configure document.
        self._document.set_style(0, 0, {
            "font_name" : self._font_name,
            "font_size" : int(self.real_font_size),
            "color" : self._color,
            "align" : "center",
        })
        # Configure layout.
        w, h = self.layout_size
        self._text_layout.width = w
        self._text_layout.height = h
        self._text_layout.content_valign = "bottom"
        self._text_layout.multiline = True
        self._text_layout.end_update()

        tl = self._text_layout


class Entry(Text):
    """Text entry.

    """

    # Keywords.

    _kwds = {
        "callback" : staticmethod(lambda text: None),
        "default" : u"",
    }

    # Constructor.

    def __init__(self, **kwds):
        super(Entry, self).__init__(**kwds)
        self._caret = _Caret(self)

    # Properties.

    def set_default(self, default):
        self._default = default
        if hasattr(self, "_caret"):
            self._caret._reset()

    def set_window(self, window):
        if self._window is not None:
            self._window.remove_handlers(self._caret)
        super(Entry, self).set_window(window)
        if self._window is not None:
            self._window.push_handlers(self._caret)


class Image(Interface):
    """Image display.

    """

    # Keywords.

    _kwds = {
        "graphic" : None,
        "callback" : staticmethod(lambda: None),
        "fitwidth" : None,
        "fitheight" : None,
    }

    _scaled = [
        ("fitwidth",),
        ("fitheight",),
    ]

    # Dimensions.

    def _content_size(self):
        gw = self.graphic.width
        gh = self.graphic.height
        lw, lh = self.layout_size
        if lw * gh > gw * lh:
            a = gw / gh
            gh = lh
            gw = lh * a
        else:
            a = gw / gh
            gw = lw
            gh = lw / a
        if self._fitwidth is not None:
            gh *= self.real_fitwidth / gw
            gw = self.real_fitwidth
        if self._fitheight is not None:
            gw *= self.real_fitheight / gh
            gh = self.real_fitheight
        return gw, gh

    # Rendering.

    def draw_content(self):
        gw = self.graphic.width
        gh = self.graphic.height
        cw, ch = self.content_size
        glScalef(cw / gw, ch / gh, 1.0)
        self.graphic.blit(0, 0)

    # Events.

    def click(self):
        make_callback(self.callback)


class Vertical(Interface):
    """Vertical layout.

    """

    # Keywords.

    _kwds = {
        "spacing" : 0.0,
    }

    _scaled = [
        (),
        ("spacing",),
    ]

    # Dimensions.

    def resize(self, width, height):
        super(Vertical, self).resize(width, height)
        self.real_spacing = self.spacing * height

    def organise(self):
        sizes = [c.content_size for c in self._components]
        widths, heights = zip(*sizes)
        max_width = max(widths)
        total_height = sum(heights)

        yoff = -self.real_spacing
        for idx in xrange(len(sizes) - 1, -1, -1):
            yoff += self.real_spacing
            xoff = (max_width - widths[idx]) * self._syms[self.halign]
            self._components[idx].reposition(xoff, yoff)
            yoff += heights[idx]

        self._content_size = lambda: (max_width, total_height)


class Tabular(Interface):
    """Tabular layout.

    """

    # Keywords.

    _kwds = {
        "spacing" : 0.0,
        "hspacing" : None,
        "vspacing" : None,
        "span" : 2,
    }

    _scaled = [
        (),
        ("spacing", "hspacing", "vspacing",),
    ]

    # Dimensions.

    def organise(self):
        hspacing = vspacing = self.real_spacing
        if self._hspacing is not None:
            hspacing = self.real_hspacing
        if self._vspacing is not None:
            vspacing = self.real_vspacing
        sizes, rows, cols = [], [], []
        for idx, component in enumerate(self._components):
            width, height = component.content_size
            rown, coln = divmod(idx, self.span)
            sizes.append((width, height))
            if rown >= len(rows): rows.append([])
            if coln >= len(cols): cols.append([])
            rows[rown].append(height)
            cols[coln].append(width)
        if self._components:
            widths = map(max, cols)
            heights = map(max, rows)
            colmax = len(widths) - 1
            rowmax = len(heights) - 1
            for idx, size in reversed(list(enumerate(sizes))):
                rown, coln = divmod(idx, self.span)
                yoff = sum(heights[rown+1:]) + vspacing * (rowmax - rown)
                xoff = sum(widths[:coln]) + hspacing * coln
                self._components[idx].reposition(xoff, yoff)
            total_width = sum(widths) + colmax * hspacing
            total_height = sum(heights) + rowmax * vspacing
            self._content_size = lambda: (total_width, total_height)


class Timer(Interface):

    _kwds = {
        "font_name" : FONT_NAME,
        "font_size" : 0.08,
    }

    _scaled = [
        (),
        ("font_size",),
    ]

    def __init__(self, **kwds):
        super(Timer, self).__init__(**kwds)
        self._alpha = 255

    def set_window(self, window):
        super(Timer, self).set_window(window)
        if self._window is None: return

        # Font and glyph array.
        self._font = pyglet.font.load(self.font_name, int(self.real_font_size))
        self._glyphs = self._font.get_glyphs(u"01234567890:")

        # Texture attributes.
        self._target = self._glyphs[0].target
        self._id = self._glyphs[0].id

        # Glyph texture coordinate data.
        self._tex_coords = [[g.tex_coords[0], g.tex_coords[1],
                             g.tex_coords[3], g.tex_coords[4],
                             g.tex_coords[6], g.tex_coords[7],
                             g.tex_coords[9], g.tex_coords[10]]
                            for g in self._glyphs]

        # Glyph vertex data.
        self._vertices = []
        a1 = self._glyphs[0].advance
        a2 = self._glyphs[-1].advance
        for glyph in self._glyphs:
            w, h = glyph.width, glyph.height
            off = (glyph.advance - w) / 2
            self._vertices.append([None] * 7)
            r = lambda x: [x - w, 0, x, 0, x, h, x - w, h]
            self._vertices[-1][0] = r(1 * a1 + 0 * a2 - off)
            self._vertices[-1][1] = r(1 * a1 + 1 * a2 - off)
            self._vertices[-1][2] = r(2 * a1 + 1 * a2 - off)
            self._vertices[-1][3] = r(3 * a1 + 1 * a2 - off)
            self._vertices[-1][4] = r(3 * a1 + 2 * a2 - off)
            self._vertices[-1][5] = r(4 * a1 + 2 * a2 - off)
            self._vertices[-1][6] = r(5 * a1 + 2 * a2 - off)

        # Report content size.
        total_width = 5 * a1 + 2 * a2
        total_height = self._glyphs[0].height
        self._content_size = lambda: (total_width, total_height)

        # Encapsulated values.
        self._minutes = 0
        self._seconds = 0
        self._hundredths = 0

        # Vertex list.
        self._vlist = pyglet.graphics.vertex_list(28,
                ('v2f', self._get_initial_vdata()),
                ('t2f', self._get_initial_tdata()),
                ('c4B', self._get_initial_cdata()))

    def _get_initial_vdata(self):
        return self._vertices[0][0] + \
               self._vertices[-1][1] + \
               self._vertices[0][2] + \
               self._vertices[0][3] + \
               self._vertices[-1][4] + \
               self._vertices[0][5] + \
               self._vertices[0][6]

    def _get_initial_tdata(self):
        return self._tex_coords[0] + \
               self._tex_coords[-1] + \
               self._tex_coords[0] + \
               self._tex_coords[0] + \
               self._tex_coords[-1] + \
               self._tex_coords[0] + \
               self._tex_coords[0]

    def _get_initial_cdata(self):
        return [255, 255, 255, 255] * 4 * 7

    def _get_alpha(self):
        return self._alpha
    def _set_alpha(self, value):
        self._alpha = value
        for i in xrange(3, len(self._vlist.colors), 4):
            self._vlist.colors[i] = value
    alpha = property(_get_alpha, _set_alpha)
    def _get_minutes(self):
        return self._minutes
    def _set_minutes(self, value):
        self._minutes = value = min(9, max(0, int(value)))
        self._vlist.vertices[0:8] = self._vertices[value][0]
        self._vlist.tex_coords[0:8] = self._tex_coords[value]
    minutes = property(_get_minutes, _set_minutes)

    def _get_seconds(self):
        return self._seconds
    def _set_seconds(self, value):
        self._seconds = value = min(59, max(0, int(value)))
        self._vlist.vertices[16:24] = self._vertices[value // 10][2]
        self._vlist.tex_coords[16:24] = self._tex_coords[value // 10]
        self._vlist.vertices[24:32] = self._vertices[value % 10][3]
        self._vlist.tex_coords[24:32] = self._tex_coords[value % 10]
    seconds = property(_get_seconds, _set_seconds)

    def _get_hundredths(self):
        return self._hundredths
    def _set_hundredths(self, value):
        self._hundredths = value = min(99, max(0, int(value)))
        self._vlist.vertices[40:48] = self._vertices[value // 10][5]
        self._vlist.tex_coords[40:48] = self._tex_coords[value // 10]
        self._vlist.vertices[48:56] = self._vertices[value % 10][6]
        self._vlist.tex_coords[48:56] = self._tex_coords[value % 10]
    hundredths = property(_get_hundredths, _set_hundredths)

    def update(self, delta):
        if delta >= 600:
            self.minutes = 9
            self.seconds = 59
            self.hundredths = 99
        else:
            seconds, fraction = divmod(delta, 1.0)
            self.minutes = seconds // 60
            self.seconds = seconds % 60
            self.hundredths = (fraction * 100) // 1

    def draw_content(self):
        glEnable(self._target)
        glBindTexture(self._target, self._id)
        glTexParameterf(self._target, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameterf(self._target, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        self._vlist.draw(GL_QUADS)
        glDisable(self._target)

class Position(Interface):

    _kwds = {
        "font_name" : FONT_NAME,
        "font_size" : 0.08,
    }

    _scaled = [
        (),
        ("font_size",),
    ]

    def __init__(self, **kwds):
        super(Position, self).__init__(**kwds)


    def set_window(self, window):
        super(Position, self).set_window(window)
        if self._window is None: return

        # Font and glyph array.
        self._font = pyglet.font.load(self.font_name, int(self.real_font_size))
        self._glyphs = self._font.get_glyphs(u"01234567890/")

        # Texture attributes.
        self._target = self._glyphs[0].target
        self._id = self._glyphs[0].id

        # Glyph texture coordinate data.
        self._tex_coords = [[g.tex_coords[0], g.tex_coords[1],
                             g.tex_coords[3], g.tex_coords[4],
                             g.tex_coords[6], g.tex_coords[7],
                             g.tex_coords[9], g.tex_coords[10]]
                            for g in self._glyphs]

        # Glyph vertex data.
        self._vertices = []
        a1 = self._glyphs[0].advance
        a2 = self._glyphs[-1].advance
        for glyph in self._glyphs:
            w, h = glyph.width, glyph.height
            off = (glyph.advance - w) / 2
            self._vertices.append([None] * 7)
            r = lambda x: [x - w, 0, x, 0, x, h, x - w, h]
            self._vertices[-1][0] = r(1 * a1 + 0 * a2 - off)
            self._vertices[-1][1] = r(2 * a1 + 0 * a2 - off)
            self._vertices[-1][2] = r(2 * a1 + 1 * a2 - off)
            self._vertices[-1][3] = r(3 * a1 + 1 * a2 - off)
            self._vertices[-1][4] = r(4 * a1 + 1 * a2 - off)
            
        # Report content size.
        total_width = 4 * a1 + 1 * a2
        total_height = self._glyphs[0].height
        self._content_size = lambda: (total_width, total_height)

        # Encapsulated values.
        self._pos = 0
        self._nracers = 0
        self._alpha = 255
        
        # Vertex list.
        self._vlist = pyglet.graphics.vertex_list(20,
                ('v2f', self._get_initial_vdata()),
                ('t2f', self._get_initial_tdata()),
                ('c4B', self._get_initial_cdata()))

    def _get_initial_vdata(self):
        return self._vertices[0][0] + \
               self._vertices[0][1] + \
               self._vertices[-1][2] + \
               self._vertices[0][3] + \
               self._vertices[0][4] 

    def _get_initial_tdata(self):
        return self._tex_coords[0] + \
               self._tex_coords[0] + \
               self._tex_coords[-1] + \
               self._tex_coords[0] + \
               self._tex_coords[0]

    def _get_initial_cdata(self):
        return [255, 255, 255, 255] * 4 * 5


    def _get_pos(self):
        return self._pos
    def _set_pos(self, value):
        self._pos = value 
        self._vlist.vertices[0:8] = self._vertices[value // 10][0]
        self._vlist.tex_coords[0:8] = self._tex_coords[value // 10]
        self._vlist.vertices[8:16] = self._vertices[value % 10][1]
        self._vlist.tex_coords[8:16] = self._tex_coords[value % 10]
    pos = property(_get_pos, _set_pos)

    def _get_nracers(self):
        return self._nracers
    def _set_nracers(self, value):
        self._nracers = value 
        self._vlist.vertices[24:32] = self._vertices[value // 10][3]
        self._vlist.tex_coords[24:32] = self._tex_coords[value // 10]
        self._vlist.vertices[32:48] = self._vertices[value % 10][4]
        self._vlist.tex_coords[32:48] = self._tex_coords[value % 10]
    nracers = property(_get_nracers, _set_nracers)

    def _get_alpha(self):
        return self._alpha
    def _set_alpha(self, value):
        self._alpha = value
        for i in xrange(3, len(self._vlist.colors), 4):
            self._vlist.colors[i] = value
    alpha = property(_get_alpha, _set_alpha)
    
    def draw_content(self):
        glEnable(self._target)
        glBindTexture(self._target, self._id)
        glTexParameterf(self._target, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameterf(self._target, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        self._vlist.draw(GL_QUADS)
        glDisable(self._target)

if __name__ == "__main__":
    import pyglet
    w = pyglet.window.Window(fullscreen=False)
    r = Root(window = w)
    t = r.add(Tabular, spacing = 0.04, halign="center", valign="center")
    Text._font_name = "Courier"
    t.add(Text, text = u"Hello", outline=(255,255,255,255))
    t.add(Text, text = u"World", outline=(255,255,255,255))
    @w.event
    def on_draw():
        w.clear()
        r.draw()
    pyglet.app.run()
