from __future__ import division

from entity import ArmedEntity
from projectile import Projectile

from vector import v
from constants import *
from common import *

from sys import _getframe

def bound(desired, limit):
    abs_limit = abs(limit)
    if desired > abs_limit:
        return abs_limit
    if desired < -abs_limit:
        return -abs_limit
    return desired
    
class Enemy(ArmedEntity):
    draw_type = 'blob_patch'
    initial_hp = 5
    recovery_time = 10
    recovery_timer = 0
    
    # some default brain parameters
    brain = None
    bp_bias = 0.0
    bp_target_theta_offset = 0.4
    bp_player_theta_offset = 0.5
    bp_turn_inaccuracy = 0
    bp_min_throttle = 0.4
    bp_firing_tolerance = 10
    bp_combat_throttle = 1.0
    bp_tangent_length = 100
    bp_catchup_distance = 0.3
    bp_can_fire = True
    
    def __init__(self, race, theta, width_offset, relative_angle, *args, **kwds):
        super(Enemy, self).__init__(race, theta, width_offset, relative_angle, *args, **kwds)
        self.fire_delay = 50
        self.accel = kwds.pop('thrust', PLAYER_THRUST)
        self.angular_accel = kwds.pop('rotation', PLAYER_ROTATION)
        
    def thrust(self, throttle = 1.0):
        self.controls['thrust_rel'] = v(self.accel, 0) * throttle
        
    def coast(self):
        self.controls['thrust_rel'] = v(0, 0)
        
    def straight_ahead(self):
        self.controls['rotation'] = 0
        
    def rotate_left(self, limit=360):
        self.controls['rotation'] = min(limit, self.angular_accel)
        
    def rotate_right(self, limit=-360):
        self.controls['rotation'] = max(limit, -self.angular_accel)

    def injure(self):
        self.recovery_timer = self.recovery_time

    @property
    def lead(self):
        if self.bp_player_theta_offset is None:
            return None
        return (self.theta - (self.race.player.theta + self.bp_player_theta_offset)) / (abs(self.bp_player_theta_offset)+0.01)
        
    def tick(self):
        if self.recovery_timer > 0:
            self.recovery_timer -= 1
        else:
            if self.brain:
                self.brain()
        super(Enemy, self).tick()
        

def follow(me, target):
    while True:
        target_vector = target.pos - me.pos
        desired_rotation = v(1, 0).rotated(me.angle).signed_angle_to(target_vector)
        me.controls['rotation'] = bound(desired_rotation, me.angular_accel)
        if abs(desired_rotation) < 15:
            me.controls['thrust_rel'] = v(bound(me.pos.distance_to(target.pos), me.accel), 0)
        else:
            me.controls['thrust_rel'] = v(0, 0)
        yield

def thrust_if_not_backward(me, track):
    racer_vec = v(1, 0).rotated(me.angle)
    if racer_vec.dot(track.tangent(me.theta)) > 0:
        l = me.lead
        if l is not None:
            me.thrust(min(1.0, max(me.bp_min_throttle, 1 - l)))
        else:
            me.thrust()
    else:
        me.coast()

def rotate_towards_direction(me, target_direction):
    racer_vec = v(1, 0).rotated(me.angle)
    side = target_direction.perpendicular().normalised().dot(racer_vec)
    if side > me.bp_bias + me.bp_turn_inaccuracy:
        me.rotate_right(limit = (me.bp_bias + me.bp_turn_inaccuracy) - side)
    elif side < me.bp_bias - me.bp_turn_inaccuracy:
        me.rotate_left(limit = (me.bp_bias + me.bp_turn_inaccuracy) - side)
    else:
        me.straight_ahead()

def aim_ahead_brain(me):
    track = me.race.track
    if me.theta < len(track.segs): 
        thrust_if_not_backward(me, track)
        fwd_vec = (track.pos(me.theta + me.bp_target_theta_offset) - me.pos)
        rotate_towards_direction(me, fwd_vec)

def sneaky_brain(me):
    if me.theta < me.race.player.theta:
        aim_ahead_brain(me)
    else:
        centre_line_brain(me)
        
def tangent_brain(me):
    track = me.race.track
    if me.theta < len(track.segs):
        thrust_if_not_backward(me, track)
        fwd_vec = track.tangent(me.theta + me.bp_target_theta_offset)
        rotate_towards_direction(me, fwd_vec)

def centre_line_brain(me):
    track = me.race.track
    if me.theta < len(track.segs):
        thrust_if_not_backward(me, track)
        fwd_vec = (track.pos(me.theta) + track.tangent(me.theta) * me.bp_tangent_length) - me.pos
        rotate_towards_direction(me, fwd_vec)

def shoot_player_brain(me):
    player_offs = me.race.player.pos - me.pos
    rotate_towards_direction(me, player_offs)
    me.thrust(me.bp_combat_throttle)
    
def harry_player_brain(fallback):
    def f(me):
        if me.theta > me.race.player.theta or me.theta + me.bp_catchup_distance < me.race.player.theta:
            fallback(me)
        else:
            shoot_player_brain(me)
    return f
        
def inside_track_brain(me):
    track = me.race.track
    if me.theta < len(track.segs):
        thrust_if_not_backward(me, track)
        target_theta = me.theta + me.bp_target_theta_offset
        curve_angle = track.normal(me.theta).signed_angle_to(track.normal(target_theta))
        if curve_angle != 0:
            target_point = track.pos(target_theta) + (track.normal(target_theta) * curve_angle).scaled_to(track.width)
        else:
            target_point = track.pos(target_theta)
        fwd_vec = (target_point - me.pos)
        rotate_towards_direction(me, fwd_vec)
             
        
#class FollowThePlayer(Enemy):
#    def __init__(self, race, theta, width_offset, angle, **kwds):
#        mybrain = follow(self, race.player)
#        super(FollowThePlayer, self).__init__(race, theta, width_offset, angle, mybrain, **kwds)
        


        
#class ShootAtPlayer(AimAtPlayer):
#    def __init__(self, race, theta, width_offset, angle, **kwds):
#        super(ShootAtPlayer, self).__init__(race, theta, width_offset, angle, **kwds)
#        
#    def tick(self):
#        super(ShootAtPlayer, self).tick()
#        if self.is_aiming_at(self.race.player):
#            self.controls['firing'] = True
#        
#    def fire(self):
#        self.race.add_projectile(Projectile(self.race, self.pos, self.theta, self.angle, self))

class EnemyRacer(Enemy):
    particles = True
    draw_type = 'svg'
    svg_name = 'usaship.svg'
    pilot_name = 'Bugs'
    
    def is_dead(self):
        return False
        
    def tick(self):
        super(EnemyRacer, self).tick()
        if self.bp_can_fire:
            player_offs = self.race.player.pos - self.pos
            da = abs(player_offs.angle - self.angle)
            self.controls['firing'] = min(da, 360 - da) < self.bp_firing_tolerance
        
class BiasRacer(EnemyRacer):
    brain = aim_ahead_brain
    bp_bias = 0.3
    svg_name = 'ussrship.svg'
    pilot_name = 'Sergei'

class AdaptiveRacer(EnemyRacer):
    brain = sneaky_brain
    bp_bias = -0.2
    svg_name = 'lgbtship.svg'
    pilot_name = 'Freedom'
            
class AngleRacer(EnemyRacer):
    brain = tangent_brain
    svg_name = 'argentineship.svg'
    pilot_name = 'Carlos'

class InsideTrackRacer(EnemyRacer):
    brain = inside_track_brain
    bp_target_theta_offset = 0.5
    svg_name = 'ukship.svg'
    pilot_name = 'Elizabeth'

class InFrontRacer(AngleRacer):
    bp_player_theta_offset = 0.05
    bp_target_theta_offset = 0.01
    svg_name = 'greeceship.svg'
    pilot_name = 'Stavros'
    
class InFrontRacer2(AngleRacer):
    bp_player_theta_offset = 0.03
    bp_target_theta_offset = 0.001    
    svg_name = 'frenchship.svg'
    pilot_name = 'Marie-Claire'
        
class CentreLineRacer(EnemyRacer):
    brain = centre_line_brain
    bp_player_theta_offset = 1.0
    svg_name = 'canadaship.svg'
    pilot_name = 'Chuck'

class CentreLineCloseInFrontRacer(CentreLineRacer):
    bp_player_theta_offset = 0.15
    svg_name = 'japanship.svg'
    pilot_name = 'Hiro'

class CentreLineInFrontRacer(CentreLineRacer):
    bp_player_theta_offset = 0.01
    svg_name = 'irishship.svg'
    pilot_name = 'Seamus'

class LoveLoveRacer(CentreLineRacer):
    bp_player_theta_offset = 0.0
    svg_name = 'usaship.svg'
    pilot_name = 'Hank'
    
class PursueRacer(EnemyRacer):
    brain = harry_player_brain(centre_line_brain)
    bp_firing_tolerance = 6
    bp_combat_throttle = 0.8
    bp_player_theta_offset = 0.2
    svg_name = 'italianship.svg'
    pilot_name = 'Giovanna'
    
class HarryRacer(EnemyRacer):
    brain = harry_player_brain(tangent_brain)
    bp_firing_tolerance = 10
    bp_combat_throttle = 0.5
    bp_player_theta_offset = 0.25
    bp_target_theta_offset = 0.2
    svg_name = 'germanship.svg'
    pilot_name = 'Dieter'
    
class AggroRacer(EnemyRacer):
    brain = harry_player_brain(tangent_brain)
    bp_player_theta_offset = 0.3
    bp_target_theta_offset = 0.2
    bp_firing_tolerance = 30
    svg_name = 'confedship.svg'
    pilot_name = 'Lee'
    
class CrazyRacer(EnemyRacer):
    brain = harry_player_brain(aim_ahead_brain)
    bp_player_theta_offset = 0.35
    bp_target_theta_offset = 0.5
    bp_bias = -0.2
    svg_name = 'jamaicaship.svg'
    pilot_name = 'Julius'

class ShootyRacer(EnemyRacer):
    brain = harry_player_brain(inside_track_brain)
    bp_target_theta_offset = 0.4
    bp_player_theta_offset = 0.3
    bp_firing_tolerance = 20
    svg_name = 'southafricaship.svg'
    pilot_name = 'Thabo'

enemy_racers = [AngleRacer,
                InsideTrackRacer,
                BiasRacer,
                CentreLineInFrontRacer,
                CentreLineCloseInFrontRacer,
                AdaptiveRacer,
                InFrontRacer,
                InFrontRacer2,
                CentreLineRacer,
                LoveLoveRacer,
                AggroRacer,
                PursueRacer,
                CrazyRacer,
                HarryRacer,
                ShootyRacer]
