from __future__ import division

import math

from vector import v, zero
from common import *

class Entity(object):
    draw_type = 'svg'
    main_svg_name = 'usaship.svg'
    ghost_svg_name = 'bwship.svg'
    radius = 24
    damage_output = 0
    initial_hp = 1
    no_collisions = False
    is_projectile = False
    particles = False
    count = 0
    def __init__(self, race, pos, **kwds):
        self.identifier = Entity.count
        Entity.count += 1
        self.race = race
        self.pos = pos
        self.old_pos = pos
        self.hp = self.initial_hp
        self.controls = {"thrust_rel": v(0, 0), "rotation": 0}
        self.collided_flag = False
        self.ghostdata = kwds.pop('ghostdata', None)
        self.is_ghost = (self.ghostdata is not None)
        if self.is_ghost:
            self.tick = self.ghost_tick
            self.interaction_mode = kwds.pop('interaction_mode', "watch")
            self.race_mode = kwds.pop('race_mode', "GP")
        
        #Could be class attribute, but we might want to upgrade it for the player.
        self.drag = 0.97
        
        #Might be better as a control. Definitely instance-mutable, though - consider player upgrades.
        self.fire_delay = 20
        self.fire_counter = 0
        
    def __cmp__(self, other):
        return cmp(self.identifier, other.identifier)
        
    @property
    def velocity(self):
        return self.pos - self.old_pos
    
    @cached
    def diameter(self):
        return 2*self.radius
    
    @cached
    def svg_name(self):
        if self.is_ghost:
            return self.ghost_svg_name
        else:
            return self.main_svg_name
    
    def is_dead(self):
        return self.hp <= 0
    
    def is_aiming_at(self, other):
        target_angle = (other.pos - self.pos).angle
        return abs(self.angle - target_angle) < 5
    
    def can_collide_with(self, other):
        return True
    
    def constrain(self):
        pass
    
    def internal_tick(self):
        self.process_controls()
        delta = (self.pos - self.old_pos) * self.drag
        self.old_pos = self.pos
        self.pos = self.pos + delta
        self.constrain()
        
    def tick(self):
        self.pos = v(floatround(self.pos[0]), floatround(self.pos[1]))
        self.old_pos = v(floatround(self.old_pos[0]), floatround(self.old_pos[1]))
        self.theta = floatround(self.theta)
        self.angle = floatround(self.angle)
        if self.race.current_tick % 50 == 0:
            #To make sure the ghost stays in synch in case float rounding does us in.
            self.collided_flag = True
        self.internal_tick()
        
    def update_ghost_controls(self, tick_key):
        if tick_key in self.ghostdata['controls']:
            self.controls = self.ghostdata['controls'][tick_key]
        for k in self.controls:
            val = self.controls[k]
            if isinstance(val, list):
                self.controls[k] = v(val)
        
    def ghost_tick_using_positions(self, tick_key):
        old_x, old_y, x, y, self.angle, self.theta = self.ghostdata['positions'][tick_key]
        self.old_pos = v(old_x / 100, old_y / 100)
        self.pos = v(x / 100, y / 100)
        
    def ghost_tick(self):
        self.pos = v(floatround(self.pos[0]), floatround(self.pos[1]))
        self.old_pos = v(floatround(self.old_pos[0]), floatround(self.old_pos[1]))
        self.theta = floatround(self.theta)
        self.angle = floatround(self.angle)
        tick_key = str(self.race.current_tick)
        if self.interaction_mode == "watch" or self.race_mode == "TT":
            self.update_ghost_controls(tick_key)
            self.internal_tick()
        else:
            self.update_ghost_controls(tick_key)
            if tick_key in self.ghostdata['positions']:
                self.ghost_tick_using_positions(tick_key)
            else:
                self.internal_tick()

    def update_controls(self):
        pass
        
    def process_controls(self):
        if 'thrust' in self.controls:
            self.pos += self.controls['thrust']
        if self.controls.get('firing'):
            if self.fire_counter <= 0:
                self.fire()
                self.fire_counter = self.fire_delay
            else:
                self.fire_counter -= 1
        else:
            self.fire_counter -= 1
            
    def cleanup(self):
        # may be called multiple times
        pass
    
    def fire(self):
        pass
        
    def injure(self):
        pass

    def hit_wall(self, bump_pos, tan):
        centre_bump_pos = bump_pos + (self.pos - bump_pos).scaled_to(self.radius)
        
        tan2 = tan.length2
        pos_off = self.pos - centre_bump_pos
        pos_ref = pos_off.dot(tan) * tan / tan2
        old_off = self.old_pos - centre_bump_pos
        old_ref = old_off.dot(tan) * tan / tan2

        norm_vel = abs(self.velocity.dot(tan.perpendicular().normalised()))
        self.old_pos = centre_bump_pos + 2 * old_ref - old_off
        self.race.dispatch_event('on_wall_bump', centre_bump_pos, self.theta, norm_vel, self)
        self.race.dispatch_event('on_bump', self, bump_pos)
        self.pos = centre_bump_pos + 2 * pos_ref - pos_off
        
    def hit(self, other):
        other.hit_by(self)
        
    def hit_by(self, other):
        bump_direction = (other.pos - self.pos)
        if bump_direction == zero:
            #Make something up
            centre_bump_pos = self.pos + v(0, 1)
            overlap_vector = v(1, 0)
        else:
            centre_bump_pos = self.pos + bump_direction.scaled_to(self.radius)
            overlap_vector = bump_direction.scaled_to(self.radius + other.radius - bump_direction.length)
        
        self_vel_orig = self.velocity
        other_vel_orig = other.velocity
        self_ratio = self.radius / (self.radius + other.radius)
        momentum = self_vel_orig * self.radius + other_vel_orig * other.radius
        mass = self.radius + other.radius
        self.pos = self.pos - overlap_vector
        other.pos = other.pos + overlap_vector
        self.old_pos = self.pos - (other_vel_orig - self_vel_orig) * (1 - self_ratio) - momentum / mass
        self.race.dispatch_event('on_bump', self, centre_bump_pos)
        self.race.dispatch_event('on_bump', other, centre_bump_pos)        
        other.old_pos = other.pos - (self_vel_orig - other_vel_orig) * self_ratio - momentum / mass

    def blast_from(self, pos, magnitude):
        offs = self.pos - pos
        if not offs.is_zero:
            force = 0.3 * magnitude / max(offs.length, 1)
            self.pos += (self.pos - pos).scaled_to(force)

    @cached            
    def finish_subtick(self):
        final_tan = self.race.track.tangent(len(self.race.track.segs))
        projected_length = final_tan.dot(self.velocity)
        if projected_length == 0: #Not sure how this happens, but it can.
            return 1
        track_end = self.race.track.pos(len(self.race.track.segs))
        projected_pos = track_end + ((self.pos - track_end).dot(final_tan) / final_tan.dot(final_tan)) * final_tan
        return min(1, max(0, (projected_pos - track_end).length / projected_length))

    
    
class TrackEntity(Entity):
    def __init__(self, race, pos, theta, **kwds):
        super(TrackEntity, self).__init__(race, pos, **kwds)
        self.theta = theta

    def constrain(self):
        if self.pos == self.old_pos: return
        segs = self.race.track.segs
        idx = int(self.theta)
        t = self.theta % 1
        candidates = []
        if idx > 0 and t < 0.1:
            candidates.append((segs[idx-1], 1.0, idx-1))
        if idx < len(segs):
            candidates.append((segs[idx], t, idx))
        if idx < len(segs) - 1 and t > 0.9:
            candidates.append((segs[idx+1], 0, idx+1))
        #assert candidates, "No candidates! Crazy! theta = %f" % self.theta
        pt = self.pos
        best_d2 = 1e100
            
        for b, guess, idx in candidates:
            t = b.closest(pt, guess)
            p = b.pos(t)
            d2 = ((p[0] - pt[0]) ** 2 + (p[1] - pt[1]) ** 2)
            if d2 < best_d2:
                best_d2 = d2
                self.theta = idx + t
                root = p
                best_t = t
                best_b = b
        #lie about the scaling in order to make collisions with the walls look better
        #on the basis that the sprites aren't actually circles.
        available_width = self.race.track.width - self.radius
        offs = pt - root
        if best_d2 < available_width ** 2:
            return
        tan = best_b.dpos(best_t)
        
        wall_pos = root + offs.scaled_to(self.race.track.width)
        self.hit_wall(wall_pos, tan)
        if (self.pos - root).length > self.race.track.width:
            self.old_pos = self.pos = root + offs.scaled_to(self.race.track.width)


class RotatingTrackEntity(TrackEntity):
    
    def __init__(self, race, pos, theta, angle, **kwds):
        super(RotatingTrackEntity, self).__init__(race, pos, theta, **kwds)
        self.angle = angle
        self.spin = 0.0
        
    def process_controls(self):
        super(RotatingTrackEntity, self).process_controls()
        self.angle += self.controls.get('rotation', 0)
        self.angle = self.controls.get('angle', self.angle)
        thrust = self.controls.get('thrust_rel', zero)
        if thrust != zero:
            self.pos += thrust.rotated(self.angle)
        if self.angle > 180: self.angle -= 360
        if self.angle < -180: self.angle += 360

    def internal_tick(self):
        super(RotatingTrackEntity, self).internal_tick()
        self.angle += self.spin
        self.spin *= 0.8
            
    def blast_from(self, pos, magnitude):
        offs = self.pos - pos
        if not offs.is_zero:
            force = 0.3 * magnitude / max(offs.length, 1)
            self.pos += (self.pos - pos).scaled_to(force)
            spin = offs.angle - self.angle
            # sine of (spin/2)
            # 3 is the magic number
            self.spin += math.sin(spin / 360 * math.pi) * 3 * magnitude / offs.length
            
#Just keeps together enemies/ghosts/players, which need better initialisation
#parameters than previously. Projectiles continue to use old system.
#TODO: This really shouldn't be part of the class structure.
class RelativePositionedEntity(RotatingTrackEntity):
    
    def __init__(self, race, theta, width_offset, relative_angle, **kwds):
        super(RelativePositionedEntity, self).__init__(
            race,
            race.pos_from_theta(theta, width_offset),
            theta,
            relative_angle + race.angle_at_theta(theta),
            **kwds)
            
class ArmedEntity(RelativePositionedEntity):
    particles = True
    finished = False
    def fire(self):
        from projectile import Projectile
        self.race.add_projectile(Projectile(self.race, self.pos, self.theta, self.angle, self))
    @property
    def no_collisions(self):
        return self.finished

class Obstacle(RelativePositionedEntity):
    svg_name = 'OBSTACLE'
    exploded = False
    def blast_from(self, pos, magnitude):
        if (pos - self.pos).length < (magnitude / 2):
            self.hp = 0
    def tick(self):
        super(Obstacle, self).tick()
        if self.hp == 0 and not self.exploded:
            self.exploded = True
            self.hp = 0
            self.race.explode_at(self.pos, self.theta, magnitude=400)
    def is_dead(self):
        return self.exploded
