"""
Model object with data for a race which is currently taking place or being simulated.

This is the 'top-level' model object.

"""

from __future__ import division

from track import generate_track
from bezier import Bezier
from vector import v
from player import Player
from enemy import *
from entity import Obstacle
from projectile import Projectile
from spatialhash import SpatialHash
from constants import *
from common import *
import pyglet
import os
import gzip
import random
try:
    import json
except:
    import simplejson as json
    
    
class Race(pyglet.event.EventDispatcher):
    event_types = ['on_entity_die', 'on_projectile_fire', 'on_bump', 'on_wall_bump', 'on_explosion']

    
    def __init__(self, name_seed, race_mode, interaction_mode, ghostdata, season=None):
        self.name = name_seed
        self.ghost_filename = name_seed + ".ghost"
        self.is_ghost = (ghostdata is not None)
        self.player_finish_time = None
        
        #Should probably avoid doing this for both the real race and the ghost race.
        self.track = generate_track(name_seed, TRACK_SEGMENTS, 200)
        
        self.entities = []
        
        self.race_mode = race_mode

        fractional_separation = EnemyRacer.radius * 14 / (5 * self.track.width)
        racer_definitions = None
        if race_mode == "GP" and (interaction_mode == "watch" or not self.is_ghost):
            grid_rng = random.Random(hash(name_seed + 'racers') & 0xffffffff)
            racers = enemy_racers[:]
            racer_definitions = []
            num_racers = grid_rng.randrange(3, len(racers))
            while num_racers > 0:
                grid_rng.shuffle(racers)
                racer_definitions.append((racers.pop(), (), {}))
                num_racers -= 1
            racer_definitions = [(Player, (), {'ghostdata':ghostdata, 'interaction_mode':interaction_mode, 'race_mode':race_mode})] + racer_definitions
            
        if season is not None and (interaction_mode == 'watch' or not self.is_ghost):
            racer_definitions = [(Player, (), {'ghostdata':ghostdata, 'interaction_mode':interaction_mode, 'race_mode':race_mode})]
            racer_definitions += map(lambda rc: (rc, (), {}), season.racers)

        if racer_definitions is not None: # GP or season mode and this is a full race
            max_per_line = self.track.width * 3 // (5 * EnemyRacer.radius)
            current_line = 0
            for ii in xrange(len(racer_definitions)):
                num_in_line = ii % max_per_line
                current_line = ii // max_per_line
                parity = 1 if num_in_line % 2 == 0 else -1
                line_parity = 1 if current_line % 2 == 0 else -1
                horizontal_position = fractional_separation * ((((num_in_line + 1) // 2) * parity) + (line_parity / 4))
                racer_cls, args, kwds = racer_definitions[ii]
                self.entities.append(racer_cls(self, current_line / 20, horizontal_position, 0, *args, **kwds))
            
            for e in self.entities:
                if isinstance(e, Player):
                    self.player = e

            self.place_obstacles(name_seed)
        elif race_mode != "TT": # GP or season mode and this is a single ghost
            self.player = Player(self, 0, fractional_separation / 4, 0, ghostdata=ghostdata, interaction_mode=interaction_mode)
            self.entities = [self.player]
        else: # time trial mode
            self.player = Player(self, 0, 0, 0, ghostdata=ghostdata, interaction_mode=interaction_mode)
            self.entities = [self.player]
 
        if DEBUG:
            self.fh = open("race.txt", "w")
        
        self.new_ghostdata = {'controls':{}, 'positions':{}, 'version':VERSION}
        self.previous_player_controls = {}
        self.current_tick = 0
        
        for r in self.racers:
            r.finished = False
        self.finishers = []

    def tick(self):
        import gc
        for e in self.entities:
            e.tick()
        
        self.handle_collisions()
        
        for e in self.entities:
            if e is not self.player and e.is_dead():
                self.remove_entity(e)
                
        for r in self.racers:
            if r.theta >= len(self.track.segs) and not r.finished:
                t = int((self.current_tick + r.finish_subtick) * 100)
                self.finishers.append((r, t))
                if r is self.player:
                    self.player_finish_time = t
                r.finished = True
        
        if (not self.is_ghost):
            if self.player.controls != self.previous_player_controls:
                self.previous_player_controls = self.player.controls.copy()
                self.new_ghostdata['controls'][self.current_tick] = self.previous_player_controls
            if self.player.collided_flag:
                self.new_ghostdata['positions'][self.current_tick] = \
                    (int(self.player.old_pos.x * 100),
                     int(self.player.old_pos.y * 100),
                     int(self.player.pos.x * 100),
                     int(self.player.pos.y * 100),
                     self.player.angle,
                     self.player.theta)
                self.player.collided_flag = False
        
        if DEBUG:    
            self.fh.write("tick = %d\n" % self.current_tick)
            for e in self.entities:
                if isinstance(e, Player) or isinstance(e, EnemyRacer):
                    self.fh.write(" %s @ %r\n" % (e.__class__.__name__, raw_values(e.pos)))    

        self.current_tick += 1
        gc.collect(0)
        
    def handle_collisions(self):
        #Who knows what a sensible cell size is?
        self.entity_hash = SpatialHash(500)
        for e in self.entities:
            if e.no_collisions: continue 
            self.entity_hash.add(e)
        collisions = set()
        for e in self.entities:
            if e.no_collisions or e.is_projectile: continue
            for target in self.entity_hash.collisions_with(e):
                if abs(target.theta - e.theta) < 1:
                    if e < target:
                        collisions.add((e, target))
                    else:
                        collisions.add((target, e))
        if not collisions: return
        collisions_list = list(collisions)
        collisions_list.sort()
        for e1, e2 in collisions_list:
            e1.hit(e2)
            
    def is_victorious(self):
        won = self.player.theta >= len(self.track.segs)
        if won:
            self.compute_finishers()
        return won
    
    def is_defeated(self):
        return self.player.is_dead()

    @cached
    def racers(self):
        return [self.player] + [x for x in self.entities if isinstance(x, EnemyRacer)]

    @property
    def player_position(self):
        if self.player.finished:
            i = 1
            for e, t in self.finishers:
                if e == self.player:
                    return i
                i += 1
                
        rl = self.racers[:]
        rl.sort(key = lambda e: e.theta, reverse = True)
        return 1 + rl.index(self.player)

    def compute_finishers(self):
        tl = len(self.track.segs)
        unfinished = [r for r in self.racers if not r.finished]
        for r in unfinished:
            r.finished = True
        new_finishers = map(lambda r: (r, int(100 * self.current_tick * tl / r.theta) + 1), unfinished)
        new_finishers.sort(key = lambda t: t[1])
        self.finishers.extend(new_finishers)
    
    def add_projectile(self, proj):
        self.entities.append(proj)
        self.dispatch_event('on_projectile_fire', proj)
        
    def remove_entity(self, e):
        e.cleanup()
        #This might be nicer if entities was a set, but so long as we don't remove entities that don't exist...
        self.entities.remove(e)
        self.dispatch_event('on_entity_die', e)
        
    def explode_at(self, pos, theta, magnitude=100):
        for o in sorted(list(self.entity_hash.objects_in_radius(pos, magnitude))):
            if abs(o.theta - theta) < 1:
                o.blast_from(pos, magnitude)
        self.dispatch_event('on_explosion', pos, magnitude)
        
    def cleanup(self):
        for e in self.entities:
            e.cleanup()
        if DEBUG:
            self.fh.close()
        
    def pos_from_theta(self, theta, width_offset = 0):
        b = self.track.segs[int(theta)]
        t = theta % 1
        root = b.pos(t)
        return root + self.track.width * width_offset * b.normal(t)
    
    def angle_at_theta(self, theta):
        b = self.track.segs[int(theta)]
        t = theta % 1
        return b.dpos(t).angle
        
    def place_obstacles(self, name_seed):
        obs_rng = random.Random(hash(name_seed + 'obstacles') & 0xffffffff)
        mode = obs_rng.choice(['drift', 'scatter'])
        if mode == 'scatter':
            current_theta = 2.0
            min_theta_diff = obs_rng.uniform(0.2, 0.5)
            max_theta_diff = min_theta_diff + obs_rng.uniform(0.0, 1.5)
            while current_theta < len(self.track.segs):
                e = Obstacle(self, current_theta, obs_rng.uniform(-0.85, 0.85), 0)
                e.pos += v(0.1,0.1)
                self.entities.append(e)
                current_theta += obs_rng.uniform(min_theta_diff, max_theta_diff)
        elif mode == 'drift':
            current_theta = 2.0
            min_drift_sep = obs_rng.uniform(0.5, 1.0)
            max_drift_sep = min_drift_sep + obs_rng.uniform(0.5, 1.5)
            obs_sep = obs_rng.uniform(0.01, 0.05)
            min_drift_count = obs_rng.randrange(3,8)
            max_drift_count = min_drift_count + obs_rng.randrange(1,6)
            while current_theta < len(self.track.segs) - max_drift_count * obs_sep:
                wmin = obs_rng.uniform(-0.9, 0.9)
                wmax = obs_rng.uniform(-0.9, 0.9)
                for n in xrange(obs_rng.randrange(min_drift_count, max_drift_count)):
                    e = Obstacle(self, current_theta, obs_rng.uniform(wmin, wmax), 0)
                    e.pos += v(0.1, 0.1)
                    self.entities.append(e)
                    current_theta += obs_rng.uniform(obs_sep, obs_sep*4)
                    if current_theta >= len(self.track.segs):
                        break
                current_theta += obs_rng.uniform(min_drift_sep, max_drift_sep)
