"""
Model object with geometry data for tracks.

Tracks should be immutable, since they could potentially be used by multiple Race
objects at once (e.g. in the case of simulating ghost data).


"""

from common import *
from vector import v, zero, LineSegment
from constants import *

from bezier import Bezier    

def generate_track(seed, min_length, max_width):
    import random
    rng = random.Random(hash(seed) & 0xffffffff)
    length = (1.00 + rng.random() * 2.00) * min_length
    width = (0.45 + rng.random() * 0.55) * max_width
    return generate_track_2(seed, length, width)

def generate_track_2(seed, length, width):
    import random
    segs = []
    start = zero
    ratio = length / TRACK_SEGMENTS
    tang = v(4*width*ratio, 0)
    rng = random.Random(hash(seed) & 0xffffffff)
    sgn = 1
    for i in xrange(TRACK_SEGMENTS):
        if rng.random() < 0.08: sgn = -sgn
        tang2 = tang.rotated(rng.uniform(sgn * 60, sgn * 30))
        end = start + tang + tang2 + tang.rotated(rng.uniform(-90, 90))
        segs.append(Bezier(start, start + tang, end - tang2, end))
        start = end
        tang = tang2
    return Track(segs, width)
        
         
class Track(object):
    def __init__(self, segs, width):
        self.segs = segs
        self.width = width
    
    def pos(self, theta):
        idx, t = divmod(theta, 1)
        idx = int(idx)
        if idx < 0: 
            idx = 0
            t = theta
        if idx >= len(self.segs): 
            idx = len(self.segs) - 1
            t = theta - idx
        return self.segs[idx].pos(t)

    def tangent(self, theta):
        idx, t = divmod(theta, 1)
        idx = int(idx)
        if idx < 0: 
            idx = 0
            t = theta
        if idx >= len(self.segs): 
            idx = len(self.segs) - 1
            t = theta - idx
        return self.segs[idx].tangent(t)

    def normal(self, theta):
        idx, t = divmod(theta, 1)
        idx = int(idx)
        if idx < 0: 
            idx = 0
            t = theta
        if idx >= len(self.segs): 
            idx = len(self.segs) - 1
            t = theta - idx
        return self.segs[idx].normal(t)
    
    def curvature(self, theta):
        idx, t = divmod(theta, 1)
        idx = int(idx)
        if idx < 0: 
            idx = 0
            t = theta
        if idx >= len(self.segs): 
            idx = len(self.segs) - 1
            t = theta - idx
        return self.segs[idx].curvature(t)
