import httplib
import functools
import pyglet
import socket
import threading
import urllib

from shared import *


## Thread
#########

class RPCThread(threading.Thread):

    def __init__(self, request, conn, timeout, name, data):
        super(RPCThread, self).__init__()
        self.daemon = True
        self.request = request
        self.conn = conn
        self.timeout = timeout
        self.name = name
        self.data = data

    def run(self):
        try:
            # Error and response states.
            error = response = None
            # Serialise the data.
            try:
                self.data = urllib.urlencode({"data": serialise(self.data)})
            except SerialiseError:
                error = RPCError(RPC_CANNOT_SERIALISE, 'input')
            # Perform the RPC request.
            try:
                if error is None:
                    self.conn.connect()
                    self.conn.sock.settimeout(self.timeout)
                    self.conn.request("POST", "/" + self.name, self.data)
            except socket.timeout, exc:
                error = RPCError(RPC_SOCKET_TIMEOUT, (exc.errno, exc.strerror))
            except socket.error, exc:
                error = RPCError(RPC_SOCKET_ERROR, (exc.errno, exc.strerror))
            # Handle the RPC response.
            if error is None:
                http_response = self.conn.getresponse()
                # Successful RPC response.
                if http_response.status == 200:
                    try:
                        response = deserialise(http_response.read())
                    except DeserialiseError:
                        error = RPCError(RPC_CANNOT_DESERIALISE, 'output')
                # Unsuccessful RPC response.
                elif http_response.status == 500:
                    error = RPCError.deserialise(http_response.reason)
                # Other response.
                else:
                    status = http_response.status
                    reason = httplib.responses[status]
                    error = RPCError(RPC_HTTP_ERROR, (status, reason))
            # Prepare request to return.
            if error is not None:
                self.request.error = error
            else:
                self.request.response = response
        # Catch unmanaged exceptions.
        except BaseException, exc:
            self.request.error = RPCError(RPC_UNKNOWN_ERROR)
            raise


## Request
##########

class RPCRequest(pyglet.event.EventDispatcher):

    event_types = ("on_rpc_error", "on_rpc_response")

    def __init__(self, *args, **kwds):
        self.thread = RPCThread(self, *args, **kwds)
        self.started = False
        self.response = None
        self.error = None

    def poll(self):
        return self.started and not self.thread.isAlive()

    def dispatch_events(self, dt=None):
        if self.poll():
            if self.error is None:
                self.dispatch_event("on_rpc_response", self, self.response)
            else:
                self.dispatch_event("on_rpc_error", self, self.error)
            return True
        return False

    def on_rpc_response(self, request, response):
        return pyglet.event.EVENT_UNHANDLED

    def on_rpc_error(self, request, error):
        print "***", error.errno,
        print ":", error.errname
        if error.errmsg:
            print "...", error.errmsg
        return pyglet.event.EVENT_UNHANDLED

    def listen(self, *args):
        self.push_handlers(*args)
        self.started = True
        self.thread.start()
        return self


## Manager
##########

class RPCManager(object):

    def __init__(self, server, port, version, debug=False, timeout=10):
        super(RPCManager, self).__init__()
        self._url = "%s:%s" % (server, port)
        self._version = tuple(version)
        self._debug = bool(debug)
        self._timeout = timeout
        self._named_rpcs = False
        self._requests = []
        self._create_named_rpcs()

    def __getattr__(self, name):
        if self._named_rpcs:
            return object.__getattribute__(self, name)
        return functools.partial(self._rpc, name)

    def dispatch_events(self, dt=None):
        for request in list(self._requests):
            if request.dispatch_events():
                self._requests.remove(request)

    def _create_named_rpcs(self):
        def on_rpc_response(request, response):
            if request == index_request:
                for name in response:
                    rpc = functools.partial(self._rpc, name)
                    setattr(self, name, rpc)
                self._named_rpcs = True
        index_request = self._rpc("server_index").listen(on_rpc_response)

    def _make_data(self, *args, **kwds):
        return {"version" : self._version,
                "devel" : self._debug,
                "args" : args,
                "kwds" : kwds}

    def _rpc(self, name, *args, **kwds):
        data = self._make_data(*args, **kwds)
        conn = httplib.HTTPConnection(self._url)
        request = RPCRequest(conn, self._timeout, name, data)
        self._requests.append(request)
        return request


## Test
#######

if __name__ == "__main__":
    # Process command line.
    import sys
    server = "snorlaxrpc.appspot.com"
    port = "80"
    if len(sys.argv) > 1:
        server = sys.argv[1]
    if len(sys.argv) > 2:
        port = sys.argv[2]
    # Create client object.
    rpm = RPCManager(server, port, (1, 0), debug=True)
    # Create live test object.
    class LiveRPC(object):
        # Handler for responses.
        def on_rpc_response(self, request, response):
            if request == self.request:
                self.response = response
        # Handler for errors.
        def on_rpc_error(self, request, error):
            if request == self.request:
                self.error = error
        def __getattr__(self, name):
            # Create RPC method.
            def liverpc(*args, **kwds):
                # Reset.
                self.request = None
                self.response = None
                self.error = None
                # Build request.
                rpc = getattr(rpm, name)
                self.request = rpc(*args, **kwds)
                self.request.listen(self)
                # Wait for a response.
                while rpm._requests:
                    rpm.dispatch_events()
                # Check for response.
                if self.response is not None:
                    response = self.response
                    return response
                # Check for error.
                if self.error is not None:
                    print "***", self.error.errno,
                    print ":", self.error.errname
                    if self.error.errmsg:
                        print "...", self.error.errmsg
            # Return wrapped RPC.
            liverpc.func_name = name
            return liverpc
    rpc = LiveRPC()
