
from __future__ import division
from collections import defaultdict

from vector import Vector

class SpatialHash(object):
    def __init__(self, cell_size):
        self.hash = defaultdict(set)
        self.cell_size = cell_size
        self.cells = {}
    def add(self, obj):
        c = self.cells_for_object(obj)
        for cell in c:
            self.hash[cell].add(obj)
        self.cells[obj] = c
        
    def remove(self, obj):
        if obj not in self.hash(obj.pos // self.cell_size):
            raise KeyError
        self.discard(obj)
        
    def discard(self, obj):
        for cell in self.cells[obj]:
            self.hash[cell].discard(obj)
            
    def collisions_with(self, obj):
        collisions = set()
        for cell in self.cells[obj]:
            for other in self.hash[cell]:
                if other is not obj:
                    if other.can_collide_with(obj) and obj.can_collide_with(other):
                        if (obj.radius + other.radius) ** 2 >= (obj.pos[0] - other.pos[0]) ** 2 + (obj.pos[1] - other.pos[1]) ** 2:
                            collisions.add(other)
        return collisions
    
    def objects_in_radius(self, pos, radius):
        collisions = set()
        for cell in self.cells_for_pos_and_radius(pos, radius):
            for other in self.hash[cell]:
                if (radius + other.radius) >= (pos - other.pos).length:
                    collisions.add(other)
        return collisions
        
    def cells_for_object(self, obj):    
        return self.cells_for_pos_and_radius(obj.pos, obj.radius)

    def cells_for_pos_and_radius(self, pos, radius):
        min_x = int((pos[0] - radius) // self.cell_size)
        max_x = int((pos[0] + radius) // self.cell_size)
        min_y = int((pos[1] - radius) // self.cell_size)
        max_y = int((pos[1] + radius) // self.cell_size)
        
        #This is clearly some kind of zip and map kind of thing.
        cells = []
        for x in xrange(min_x, max_x + 1):
            for y in xrange(min_y, max_y + 1):
                cells.append((x, y))
        return cells
    




#import unittest
#
#class Obj(object):
#    def __init__(self, pos, radius):
#        self.pos = Vector(pos)
#        self.radius = radius
#        
#
#class TestSpatialHash(unittest.TestCase):
#    
#    def setUp(self):
#        self.cell5 = SpatialHash(5)
#        self.cell10 = SpatialHash(10)
#        
#    def testNoSelfCollision(self):
#        obj = Obj((14, 32), 7)
#        self.cell5.add(obj)
#        self.assertEqual([], self.cell5.collisions_with(obj))
#        
#    def testNoCollision(self):
#        self.cell5.add(Obj((14, 32), 7))
#        self.assertEqual([], self.cell5.collisions_with(Obj((500, 500), 1)))
#        
#    def testBasicCollision(self):
#        obj = Obj((5, 5), 2)
#        self.cell10.add(obj)
#        self.assertEqual([obj], self.cell10.collisions_with(Obj((6, 6), 2)))
#        
#    def testMultipleCollisions(self):
#        obj1 = Obj((5, 5), 4)
#        obj2 = Obj((25, 25), 4)
#        self.cell10.add(obj1)
#        self.cell10.add(obj2)
#        self.assertEqual([obj1, obj2], self.cell10.collisions_with(Obj((15, 15), 11)))
#        
#
#
#if __name__ == '__main__':
#    unittest.main()
    
    
    
    
    